import sys
import argparse
import pickle
import subprocess
from ..analysis import analyse
from ..scenarios import run_scenarios
from ..reconstruction import reconstruct


def main():
    argp = argparse.ArgumentParser(
        description="Disasemble binaries and recover as much info as possible"
    )
    argp.add_argument("file", type=str, help="File to dis-cover")
    argp.add_argument(
        "-d",
        "--output-directory",
        type=str,
        default="/tmp",
        help='Directory where the temporary files are written (default "/tmp")',
    )
    argp.add_argument(
        "-p",
        "--pickle",
        action="store_true",
        default=False,
        help="Output info in the pickle format (used with --bin)",
    )
    argp.add_argument(
        "-o",
        "--output-file",
        type=str,
        default="./reconstructed",
        help='File where the output should be written (used with --bin) (default "./reconstructed")',
    )
    argp.add_argument(
        "-l",
        "--list-classes",
        action="store_true",
        default=False,
        help="List the classes found in the binary (used with --bin)",
    )
    group = argp.add_mutually_exclusive_group()
    group.add_argument(
        "-b",
        "--bin",
        action="store_true",
        default=True,
        help="Extract info from a binary file (default)",
    )
    group.add_argument(
        "-c",
        "--cpp",
        action="store_true",
        help="Compile C++ file under multiple scenarios and extract info from the given outputs",
    )
    arguments = argp.parse_args()

    if arguments.cpp:
        run_scenarios(arguments.file, arguments.output_directory)
    # It's important to check this last, as it defaults to True
    elif arguments.bin:
        analysis = analyse(arguments.file)
        if arguments.pickle:
            sys.stdout.buffer.write(pickle.dumps(analysis.classes))
        else:
            if arguments.list_classes:
                print(analysis)
            else:
                print(
                    "🔎  Analysis has found %d classes in %s"
                    % (len(analysis.get_classes()), arguments.file)
                )
            reconstruction = reconstruct(analysis)
            file_name = arguments.file.split("/")[-1]
            reconstructed_file_path = (
                arguments.output_directory + "/" + file_name + "_reconstructed"
            )
            print("🏗️   Writing reconstructed ELF to %s" % reconstructed_file_path)
            reconstructed_file = open(reconstructed_file_path, "wb")
            reconstructed_file.write(reconstruction)
            reconstructed_file.close()
            if not check_for_command("objcopy"):
                return
            stripped_file_path = (
                arguments.output_directory + "/" + file_name + "_stripped"
            )
            print("✂️   Stripping original ELF to %s" % stripped_file_path)
            strip = subprocess.run(
                ["objcopy", "--strip-all", arguments.file, stripped_file_path]
            )
            if not check_for_command("eu-unstrip"):
                return
            print(
                "🪢   Combining reconstructed and stripped ELFs to %s"
                % arguments.output_file
            )
            unstrip = subprocess.run(
                [
                    "eu-unstrip",
                    stripped_file_path,
                    reconstructed_file_path,
                    "-o",
                    arguments.output_file,
                ]
            )


# Check if command exists on the system
def check_for_command(command):
    c = subprocess.run("command -v " + command, shell=True, capture_output=True)
    if c.returncode == 0:
        return 1
    else:
        print(command + " is needed to continue with the process")
        return 0



import struct
from elftools.elf.elffile import ELFFile
from itanium_demangler import parse as demangle


DATA_SECTIONS = [".rodata", ".data.rel.ro", ".data.rel.ro.local", ".rdata"]


class CppClass:
    def __init__(self, name):
        self.name = name
        self.inherits_from = set()
        self.address = None

    def __str__(self):
        output = self.name
        if len(self.inherits_from) > 0:
            output += "\tinherits from %s" % ", ".join(self.inherits_from)
        return output

    def __hash__(self):
        return hash(
            str(hash(self.name))
            + str(hash("".join(sorted([str(hash(c)) for c in self.inherits_from]))))
        )

    def __eq__(self, other):
        return hash(self) == hash(other)


class ElfAnalysis:
    """An analysis of an ELF file"""

    def __init__(self, file_name):
        self.file_name = file_name
        f = open(self.file_name, "rb")
        self.elffile = ELFFile(f)
        self.sections = []
        for section in self.elffile.iter_sections():
            self.sections.append(
                (section["sh_addr"] + section["sh_size"], section.name)
            )
        self.sections.sort(key=lambda i: i[0])
        self.classes = []
        self.names = {}
        self.sections_data = {}
        self.addresses = []

    def get_section_name(self, addr):
        if addr < 0:
            return "out of bounds"
        for (end, name) in self.sections:
            if addr < end:
                return name
        return "out of bounds"

    def get_section_data_by_name(self, name):
        if self.sections_data.get(name) == None:
            self.sections_data[name] = self.elffile.get_section_by_name(name).data()
        return self.sections_data[name]

    def extract_name(self, addr):
        if self.names.get(addr) != None:
            return self.names[addr]
        section_name = self.get_section_name(addr)
        if section_name == "out of bounds":
            return
        section = self.elffile.get_section_by_name(section_name)
        section_data = self.get_section_data_by_name(section_name)
        relative_address = addr - section["sh_addr"]
        if relative_address >= 0 and relative_address < len(section_data):
            name = "_Z"
            while (
                relative_address < len(section_data)
                and section_data[relative_address] != 0
            ):
                name += chr(section_data[relative_address])
                relative_address += 1
            self.names[addr] = str(demangle(name))
            try:
                return str(demangle(name))
            except:
                return str(name)

    def __str__(self):
        output = "🔎  List of classes found in %s:\n" % self.file_name
        return output + "\n".join([str(c) for c in self.get_classes()])

    def get_classes(self):
        return self.classes

    def extract_rtti_info(self):

        self.program_map = {}

        for data_section_name in DATA_SECTIONS:

            data_section = self.elffile.get_section_by_name(data_section_name)

            # If there is no section, we go to the next one
            if not data_section:
                continue

            data = self.get_section_data_by_name(data_section_name)

            base_address = data_section["sh_addr"]

            for offset in range(0, len(data), 8):
                line = list(data[offset : offset + 8])
                line.reverse()
                line_str = "".join([format(d, "02x") for d in line])
                line_int = int(line_str, 16)
                section = self.get_section_name(line_int)

                flag = "unknown"

                if section in DATA_SECTIONS:
                    flag = "data"
                elif line_int == 0:
                    flag = "zeroes"

                self.program_map[base_address + offset] = (line_int, flag)

        self.addresses = list(self.program_map.keys())
        self.addresses.sort()

        # Add "begin_vtable" and "begin_rtti" flags
        # Append CppClass objects to self.classes
        for address in self.addresses:
            (line, flag) = self.program_map[address]
            next_v = self.program_map.get(address + 8)
            if not next_v:
                continue
            (next_line, next_flag) = next_v
            # First we find out if this is the beginning of a vtable
            if flag == "zeroes" and next_flag == "data":
                # If it is, we find the associated RTTI
                success, cpp_class = self.flag_rtti_recur(next_line)
                if cpp_class:
                    cpp_class.vtable_address = address
                # If we have successfuly flagged an RTTI, then at
                # address there is a vtable
                if success:
                    self.program_map[address] = (line, "begin_vtable")

    def flag_rtti_recur(self, address):
        # We check that the beginning of the table is the beginning of an RTTI
        rtti_start = self.program_map.get(address)
        if not rtti_start:
            return False, None
        (line, flag) = rtti_start

        if flag == "begin_rtti":
            name = self.extract_name(self.program_map[address + 8][0])
            for cpp_class in self.classes:
                if cpp_class.name == name:
                    return name, cpp_class
            return name, None

        if flag not in ["unknown", "zeroes"]:
            return False, None

        # We check that the next part of the table is a name
        name_field = self.program_map.get(address + 8)
        if not name_field:
            return False, None
        (name_line, name_flag) = name_field
        if name_flag != "data":
            return False, None

        name = self.extract_name(name_line)
        if not name or name == "None":
            return False, None

        cpp_class = CppClass(name)
        cpp_class.address = address
        self.program_map[address] = (line, "begin_rtti")

        i = 2
        while address + 8 * i in self.addresses:
            (parent_line, parent_flag) = self.program_map[address + 8 * i]
            i += 1
            if parent_flag in ["zeroes", "begin_rtti", "begin_vtable"]:
                break
            parent_name, _ = self.flag_rtti_recur(parent_line)
            if not parent_name:
                continue
            cpp_class.inherits_from.add(parent_name)

        self.classes.append(cpp_class)

        return name, cpp_class


def analyse(elf_file_name):

    analysis = ElfAnalysis(elf_file_name)

    analysis.extract_rtti_info()

    return analysis



from elftools.elf.enums import ENUM_SH_TYPE_BASE, ENUM_P_TYPE_BASE

SECTIONS_TO_KEEP = [
    ".note.gnu.build-id",
    ".note.ABI-tag",
    ".comment",
    ".dynsym",
]

SECTIONS_TO_CREATE = [
    ".debug_info",  # main dwarf info
    ".debug_abbrev",  # dwarf type definition
    ".debug_line",  # source code line info
    ".debug_str",  # string table for debug info
    ".debug_ranges",  # address ranges
    ".symtab",  # symbolic table
    ".strtab",  # string table for symtab
]

DEBUG_SECTION_HEADER = {
    "sh_type": "SHT_PROGBITS",
    "sh_flags": 0,
    "sh_addr": 0,
    "sh_link": 0,
    "sh_info": 0,
    "sh_addralign": 1,
    "sh_entsize": 0,
}

SYMTAB_SECTION_HEADER = {
    "sh_type": "SHT_SYMTAB",
    "sh_flags": 0,
    "sh_addr": 0,
    "sh_link": 0,  # Modified in the code to .strtab's index
    "sh_info": 0,
    "sh_addralign": 8,
    "sh_entsize": 24,
}


def build_symtab_section_header(sh_link):
    SYMTAB_SECTION_HEADER["sh_link"] = sh_link
    return SYMTAB_SECTION_HEADER


STRTAB_SECTION_HEADER = {
    "sh_type": "SHT_STRTAB",
    "sh_flags": 0,
    "sh_addr": 0,
    "sh_link": 0,
    "sh_info": 0,
    "sh_addralign": 1,
    "sh_entsize": 0,
}

# Transforms an int to a bytes representation, in little endian format, padded
# using the width argument (in bytes).
def int_to_bytes(i, width=2):
    return bytes.fromhex(hex(i)[2:].zfill(width * 2))[::-1]


class Reconstruction:
    """The reconstruction after an analysis"""

    def __init__(self, analysis):
        self.elffile = analysis.elffile
        self.classes = analysis.classes

    def reconstruct(self):
        # The four main parts of an ELF file that we will fill out
        self.elf_header = b""
        self.program_header_table = b""
        self.sections = b""
        self.section_header_table = b""

        self.construct_program_header_table()
        self.construct_sections_and_sections_header_table()
        self.construct_elf_header()

        self.data = b""
        self.data += self.elf_header
        self.data += self.program_header_table
        self.data += self.sections
        self.data += self.section_header_table

    # Construct the program header table
    def construct_program_header_table(self):
        # We simply copy the original header table, with some p_filesz fields set
        # to 0
        for segment in self.elffile.iter_segments():
            header = segment.header
            program_header = b""
            program_header += int_to_bytes(ENUM_P_TYPE_BASE[header["p_type"]], width=4)
            program_header += int_to_bytes(header["p_flags"], width=4)
            program_header += int_to_bytes(header["p_offset"], width=8)
            program_header += int_to_bytes(header["p_vaddr"], width=8)
            program_header += int_to_bytes(header["p_paddr"], width=8)
            if segment.header["p_type"] in ["PT_PHDR", "PT_NOTE"]:
                program_header += int_to_bytes(header["p_filesz"], width=8)
            else:
                program_header += int_to_bytes(0, width=8)
            program_header += int_to_bytes(header["p_memsz"], width=8)
            program_header += int_to_bytes(header["p_align"], width=8)

            self.program_header_table += program_header

    # Construct the sections and section header table
    def construct_sections_and_sections_header_table(self):

        self.sections_list = []
        self.e_shnum = 0
        self.e_shstrndx = 0
        self.shstrtab_data = b""
        self.sections_offset = int("0x40", 16) + len(self.program_header_table)
        # self.sections_offset = len(self.program_header_table)

        # The first field is always empty
        # self.section_header_table += b"\x00" * 64

        for section in self.elffile.iter_sections():
            section_data = b""
            section_header = b""

            # We skip the .shstrtab section, we will do it after the SECTIONS_TO_CREATE
            if section.name == ".shstrtab":
                continue
            # We skip these for now if they exist
            elif section.name in SECTIONS_TO_CREATE:
                continue

            if section.name == ".data.rel.ro":
                self.st_shndx = self.e_shnum

            self.e_shnum += 1

            # All of the sections we keep as is
            if section.name in SECTIONS_TO_KEEP:
                section_data += section.data()
                section_header += self.create_section_header(section.header)
            # The other sections (we squash them)
            else:
                # TODO set SHT_NOBITS flag
                section_data += bytes(section.header["sh_size"])
                section_header += self.create_section_header(section.header)

            # We add the name to the shstrtab section
            self.shstrtab_data += section.name.encode() + b"\x00"
            self.sections += section_data
            self.section_header_table += section_header
            if section.name:
                self.sections_list.append(section.name)

        # We build the debug sections to add them to the sections later
        self.build_debug_sections()
        # We build the table sections to add them to the sections later
        self.build_table_sections()

        # All of the sections we create
        for section_name in SECTIONS_TO_CREATE:
            header = DEBUG_SECTION_HEADER
            self.e_shnum += 1
            section_header = b""
            section_data = b""
            if section_name == ".debug_info":
                section_data += self.debug_info
            elif section_name == ".debug_abbrev":
                section_data += self.debug_abbrev
            elif section_name == ".debug_str":
                section_data += self.debug_str
            elif section_name == ".symtab":
                section_data += self.symtab
                header = build_symtab_section_header(self.e_shnum)
            elif section_name == ".strtab":
                section_data += self.strtab
                header = STRTAB_SECTION_HEADER
            section_header += self.create_section_header(
                header,
                sh_size=len(section_data),
            )
            self.shstrtab_data += section_name.encode() + b"\x00"
            self.sections += section_data
            self.section_header_table += section_header
            self.sections_list.append(section.name)

        # We write the .shstrtab section
        self.e_shstrndx = self.e_shnum
        self.e_shnum += 1
        shstrtab_name = b".shstrtab\x00"
        self.section_header_table += self.create_section_header(
            self.elffile.get_section_by_name(".shstrtab").header,
            sh_size=len(self.shstrtab_data) + len(shstrtab_name),
        )
        self.shstrtab_data += shstrtab_name
        self.sections += self.shstrtab_data
        self.sections_list.append(".shstrtab")

    def create_section_header(self, header, sh_size=-1):
        if sh_size == -1:
            sh_size = header["sh_size"]
        section_header = b""
        section_header += int_to_bytes(len(self.shstrtab_data), width=4)
        section_header += int_to_bytes(ENUM_SH_TYPE_BASE[header["sh_type"]], width=4)
        section_header += int_to_bytes(header["sh_flags"], width=8)
        section_header += int_to_bytes(header["sh_addr"], width=8)
        section_header += int_to_bytes(
            self.sections_offset + len(self.sections), width=8
        )
        section_header += int_to_bytes(sh_size, width=8)
        section_header += int_to_bytes(header["sh_link"], width=4)
        section_header += int_to_bytes(header["sh_info"], width=4)
        section_header += int_to_bytes(header["sh_addralign"], width=8)
        section_header += int_to_bytes(header["sh_entsize"], width=8)
        return section_header

    def build_debug_sections(self):
        debug_info = b""
        debug_abbrev = b""
        debug_str = b""

        # We create the debug_abbrev types

        debug_abbrev += b"\x01"  # abbrev 1
        debug_abbrev += b"\x11"  # compile_unit
        debug_abbrev += b"\x01"  # has children
        debug_abbrev += b"\x13\x05"  # language: data2
        debug_abbrev += b"\x11\x01"  # low_pc: address
        debug_abbrev += b"\x55\x17"  # ranges: sec_offset
        debug_abbrev += b"\x00\x00"

        debug_abbrev += b"\x02"  # abbrev 2
        debug_abbrev += b"\x02"  # class_type
        debug_abbrev += b"\x01"  # has children
        debug_abbrev += b"\x1d\x13"  # containing_type: ref4
        debug_abbrev += b"\x36\x0b"  # calling_convention: data1 EXPANDABLE ?
        debug_abbrev += b"\x03\x0e"  # name: strp
        debug_abbrev += b"\x0b\x0b"  # byte_size: data1
        debug_abbrev += b"\x00\x00"

        debug_abbrev += b"\x03"  # abbrev 3
        debug_abbrev += b"\x1c"  # inheritance
        debug_abbrev += b"\x00"  # no children
        debug_abbrev += b"\x49\x13"  # type: ref4
        debug_abbrev += b"\x00\x00"

        debug_abbrev += b"\x00"

        # We create the compilation unit header
        # The first field, unit_length, is added at the end of this method
        debug_info += int_to_bytes(4)  # version
        debug_info += bytes(4)  # debug_abbrev_offset
        debug_info += int_to_bytes(8, width=1)  # address_size

        # The root DIE
        debug_info += b"\x01"  # abbrev number
        debug_info += b"\x21\x00"  # language (C++)
        # TODO figure out which is 8byte and which is 4
        # (in theory, both are 8)
        debug_info += b"\x00\x00\x00\x00\x00\x00\x00\x00"  # low_pc
        debug_info += b"\x00\x00\x00\x00"  # ranges

        self.first_class_location = len(debug_info) + 4

        for cpp_class in self.classes:
            # Class
            type_location = (
                len(debug_info) + 4
            )  # TODO find out what happens with multiple inheritance for the containing type
            debug_info += b"\x02"  # abbrev number
            debug_info += int_to_bytes(type_location, width=4)  # containing type
            debug_info += b"\x04"  # calling_convention
            debug_info += int_to_bytes(len(debug_str), width=4)  # name
            debug_info += (
                b"\x08"  # byte_size # TODO maybe find out the size of the class
            )

            # We add the name to debug_str
            debug_str += cpp_class.name.encode() + b"\x00"

            # Inherits from
            for inheritance in cpp_class.inherits_from:
                debug_info += b"\x03"  # abbrev number
                debug_info += int_to_bytes(
                    self.find_class_location(inheritance), width=4
                )  # type

            # Null tag
            debug_info += b"\x00"

        # Null tag
        debug_info += b"\x00"

        # Finally, we add the unit_length (initial length) at the beginning of the
        # compilation unit header
        debug_info = int_to_bytes(len(debug_info), width=4) + debug_info

        self.debug_info = debug_info
        self.debug_abbrev = debug_abbrev
        self.debug_str = debug_str

    def find_class_location(self, cpp_class):
        location = self.first_class_location
        for c in self.classes:
            if c.name == cpp_class:
                return location
            location += 12
            for i in c.inherits_from:
                location += 5
        return 0

    def build_table_sections(self):

        # First, we check the existing symtab and import the existing symbols
        symtab_section = self.elffile.get_section_by_name(".symtab")
        n = -1

        symtab = b""
        strtab = b""

        symtab += b"\x00" * 24
        strtab += b"\x00"

        if symtab_section:
            for symbol in symtab_section.iter_symbols():
                n += 1
                skip_symbol = False

                # We check if the type of the table is "STT_SECTION"
                if symbol["st_info"]["type"] == "STT_SECTION":
                    continue

                # We check if the symbol corresponds to a class we have discovered
                for cpp_class in self.classes:
                    name = cpp_class.name
                    rtti_name = "_ZTI" + mangle(name)
                    vtable_name = "_ZTV" + mangle(name)
                    # If it is, we skip it to add it later
                    if symbol.name in [rtti_name, vtable_name]:
                        skip_symbol = True
                        break

                if skip_symbol:
                    continue

                entry_offset = (
                    symtab_section["sh_offset"] + n * symtab_section["sh_entsize"]
                )
                symtab_section.stream.seek(entry_offset)
                symbol_value = symtab_section.stream.read(symtab_section["sh_entsize"])

                # We do not include the symbol if it is null
                if symbol_value == b"\x00" * symtab_section["sh_entsize"]:
                    continue

                # We find the id of the section pointed to by the symbol
                st_shndx = symbol["st_shndx"]
                if isinstance(st_shndx, int):
                    section_name = self.elffile.get_section(st_shndx).name
                    try:
                        new_st_shndx = self.sections_list.index(section_name)
                    except:
                        continue

                    # Now we can create the symbol with new st_name and st_shndx
                    symtab += int_to_bytes(len(strtab), width=4)  # st_name
                    symtab += symbol_value[4:6]  # st_info, st_other
                    symtab += int_to_bytes(new_st_shndx, width=2)  # st_shndx
                    symtab += symbol_value[8:]  # st_value, st_size

                    strtab += symbol.name.encode() + b"\x00"

        for c in self.classes:
            symtab += int_to_bytes(len(strtab), width=4)  # st_name
            symtab += b"\x21"  # st_info
            symtab += b"\x00"  # st_other
            symtab += int_to_bytes(self.st_shndx, width=2)  # st_shndx (.data.rel.ro id)
            symtab += int_to_bytes(c.address, width=8)  # st_value
            symtab += b"\x10\x00\x00\x00\x00\x00\x00\x00"  # st_size # TODO

            strtab += b"_ZTI" + mangle(c.name).encode() + b"\x00"

            symtab += int_to_bytes(len(strtab), width=4)  # st_name
            symtab += b"\x21"  # st_info
            symtab += b"\x00"  # st_other
            symtab += int_to_bytes(self.st_shndx, width=2)  # st_shndx (.data.rel.ro id)
            try:
                symtab += int_to_bytes(c.vtable_address, width=8)  # st_value
            except:
                symtab += b"\x00" * 8  # st_value
            symtab += b"\x20\x00\x00\x00\x00\x00\x00\x00"  # st_size # TODO

            strtab += b"_ZTV" + mangle(c.name).encode() + b"\x00"

        self.symtab = symtab
        self.strtab = strtab

    def construct_elf_header(self):
        # We construct the elf header
        self.elffile.stream.seek(0)
        self.elf_header += self.elffile.stream.read(
            int("0x20", 16)
        )  # Read until e_phoff

        self.elf_header += int_to_bytes(int("0x40", 16), width=8)  # e_phoff
        self.elf_header += int_to_bytes(
            int("0x40", 16) + len(self.program_header_table) + len(self.sections),
            width=8,
        )  # e_shoff

        self.elffile.stream.seek(int("0x30", 16))  # Seek at e_flags
        self.elf_header += self.elffile.stream.read(12)  # Read until e_shnum

        self.elf_header += int_to_bytes(self.e_shnum)  # e_shnum
        self.elf_header += int_to_bytes(self.e_shstrndx)  # e_shstrndx


# Mangle a name according to the abi
def mangle(name):
    namespaces = name.split("::")
    mangled_name = ""
    for namespace in namespaces:
        mangled_name += str(len(namespace)) + namespace
    if len(namespaces) > 1:
        mangled_name = "N" + mangled_name + "E"
    return mangled_name


def reconstruct(analysis):

    reconstruction = Reconstruction(analysis)

    reconstruction.reconstruct()

    return reconstruction.data



import subprocess


def compile_under_scenario(source_file_name, scenario, output_directory):
    (compiler, option) = scenario
    output_file_name = "%s/%s_%s_%s" % (
        output_directory,
        source_file_name.split("/")[-1][:-4],
        compiler,
        option[1:],
    )
    command = [
        compiler,
        "-pie",
        "-fPIC",
        "-fPIE",
        option,
        source_file_name,
        "-o",
        output_file_name,
    ]
    """
    print(
        "Compilation details : \033[1m%s\033[0m and \033[1m%s\033[0m"
        % (compiler, option)
    )
    """
    process = subprocess.run(command)
    if process.returncode == 0:
        return output_file_name
